/*
  FUSE: Filesystem in Userspace
  Copyright (C) 2001-2007  Miklos Szeredi <miklos@szeredi.hu>
  Copyright (C) 2011       Sebastian Pipping <sebastian@pipping.org>

  This program can be distributed under the terms of the GNU GPL.
  See the file COPYING.

  gcc -Wall fuseproxy_fh.c `pkg-config fuse --cflags --libs` -lulockmgr -o fuseproxy_fh
*/


#include "spfs_config.h"

#include <fuse.h>
#include <ulockmgr.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include <dirent.h>
#ifdef HAVE_SETXATTR
#include <sys/xattr.h>
#endif
#include <sys/file.h> /* flock(2) */

#include "include/util.h"
#include "include/log.h"

#define EINTR_TRIES	3

static int proxy_getattr(const char *path, struct stat *stbuf)
{
	int res;

	res = lstat(path, stbuf);
	if (res == -1)
		return -errno;

	return 0;
}

static int proxy_fgetattr(const char *path, struct stat *stbuf,
			struct fuse_file_info *fi)
{
	int res;

	(void) path;

	res = fstat(fi->fh, stbuf);
	if (res == -1)
		return -errno;

	return 0;
}

static int proxy_access(const char *path, int mask)
{
	int res;

	res = access(path, mask);
	if (res == -1)
		return -errno;

	return 0;
}

static int proxy_readlink(const char *path, char *buf, size_t size)
{
	int res;

	res = readlink(path, buf, size - 1);
	if (res == -1)
		return -errno;

	buf[res] = '\0';
	return 0;
}

struct proxy_dirp {
	DIR *dp;
	struct dirent *entry;
	off_t offset;
};

static int proxy_opendir(const char *path, struct fuse_file_info *fi)
{
	int res;
	struct proxy_dirp *d = malloc(sizeof(struct proxy_dirp));
	if (d == NULL)
		return -ENOMEM;

	d->dp = opendir(path);
	if (d->dp == NULL) {
		res = -errno;
		free(d);
		return res;
	}
	d->offset = 0;
	d->entry = NULL;

	pr_debug("%s: opened %s as fd %d\n", __func__, path, dirfd(d->dp));
	fi->fh = (unsigned long) d;
	return 0;
}

static inline struct proxy_dirp *get_dirp(struct fuse_file_info *fi)
{
	return (struct proxy_dirp *) (uintptr_t) fi->fh;
}

static int proxy_readdir(const char *path, void *buf, fuse_fill_dir_t filler,
		       off_t offset, struct fuse_file_info *fi)
{
	struct proxy_dirp *d = get_dirp(fi);

	(void) path;
	if (offset != d->offset) {
		seekdir(d->dp, offset);
		d->entry = NULL;
		d->offset = offset;
	}
	while (1) {
		struct stat st;
		off_t nextoff;

		if (!d->entry) {
			d->entry = readdir(d->dp);
			if (!d->entry)
				break;
		}

		memset(&st, 0, sizeof(st));
		st.st_ino = d->entry->d_ino;
		st.st_mode = d->entry->d_type << 12;
		nextoff = telldir(d->dp);
		if (filler(buf, d->entry->d_name, &st, nextoff))
			break;

		d->entry = NULL;
		d->offset = nextoff;
	}

	return 0;
}

static int proxy_releasedir(const char *path, struct fuse_file_info *fi)
{
	struct proxy_dirp *d = get_dirp(fi);

	pr_debug("%s: closed %s as fd %d\n", __func__, path, dirfd(d->dp));

	(void) path;
	closedir(d->dp);
	free(d);
	return 0;
}

static int proxy_mknod(const char *path, mode_t mode, dev_t rdev)
{
	int res;

	if (S_ISFIFO(mode))
		res = mkfifo(path, mode);
	else
		res = mknod(path, mode, rdev);
	if (res == -1)
		return -errno;

	return 0;
}

static int proxy_mkdir(const char *path, mode_t mode)
{
	int res;

	res = mkdir(path, mode);
	if (res == -1)
		return -errno;

	return 0;
}

static int proxy_unlink(const char *path)
{
	int res;

	res = unlink(path);
	if (res == -1)
		return -errno;

	return 0;
}

static int proxy_rmdir(const char *path)
{
	int res;

	res = rmdir(path);
	if (res == -1)
		return -errno;

	return 0;
}

static int proxy_symlink(const char *from, const char *to)
{
	int res;

	res = symlink(from, to);
	if (res == -1)
		return -errno;

	return 0;
}

static int proxy_rename(const char *from, const char *to)
{
	int res;

	pr_debug("%s: rename %s to %s\n", __func__, from, to);

	res = rename(from, to);
	if (res == -1)
		return -errno;

	return 0;
}

static int proxy_link(const char *from, const char *to)
{
	int res;

	res = link(from, to);
	if (res == -1)
		return -errno;

	return 0;
}

static int proxy_chmod(const char *path, mode_t mode)
{
	int res;

	res = chmod(path, mode);
	if (res == -1)
		return -errno;

	return 0;
}

static int proxy_chown(const char *path, uid_t uid, gid_t gid)
{
	int res;

	res = lchown(path, uid, gid);
	if (res == -1)
		return -errno;

	return 0;
}

static int proxy_truncate(const char *path, off_t size)
{
	int res;
	int tries = 0;

	while (tries++ < EINTR_TRIES) {
		res = truncate(path, size);
		if (res == -1) {
			if (errno == EINTR) {
				pr_warn("%s: truncate returned -EINTR; retrying\n", __func__);
				continue;
			}
			return -errno;
		}

		return 0;
	}
	pr_warn("%s: failed; returning -EINTR\n", __func__);
	return -EINTR;
}

static int proxy_ftruncate(const char *path, off_t size,
			 struct fuse_file_info *fi)
{
	int res;
	int tries = 0;

	(void) path;

	while (tries++ < EINTR_TRIES) {
		res = ftruncate(fi->fh, size);
		if (res == -1) {
			if (errno == EINTR) {
				pr_warn("%s: ftruncate returned -EINTR; retrying\n", __func__);
				continue;
			}
			return -errno;
		}

		return 0;
	}
	pr_warn("%s: failed; returning -EINTR\n", __func__);
	return -EINTR;
}

#ifdef HAVE_UTIMENSAT
static int proxy_utimens(const char *path, const struct timespec ts[2])
{
	int res;

	/* don't use utime/utimes since they follow symlinks */
	res = utimensat(0, path, ts, AT_SYMLINK_NOFOLLOW);
	if (res == -1)
		return -errno;

	return 0;
}
#endif

static int proxy_create(const char *path, mode_t mode, struct fuse_file_info *fi)
{
	int fd;
	int tries = 0;

	pr_debug("%s: fi->flags: %o\n", __func__, fi->flags);
	pr_debug("%s: mode: %o\n", __func__, mode);

	while (tries++ < EINTR_TRIES) {
		fd = open(path, fi->flags, mode);
		if (fd == -1) {
			if (errno == EINTR) {
				pr_warn("%s: %s returned -EINTR; retrying\n",
						__func__,
						mode ? "create" : "open");
				continue;
			}
			return -errno;
		}

		pr_debug("%s: success\n", __func__);

		fi->fh = fd;
		return 0;
	}
	pr_warn("%s: failed; returning -EINTR\n", __func__);
	return -EINTR;
}

static int proxy_open(const char *path, struct fuse_file_info *fi)
{
	return proxy_create(path, 0, fi);
}

static int proxy_read(const char *path, char *buf, size_t size, off_t offset,
		    struct fuse_file_info *fi)
{
	ssize_t res;
	int tries = 0;

	(void) path;

	while (tries++ < EINTR_TRIES) {
		errno = 0;
		res = pread(fi->fh, buf, size, offset);
		if (res == -1) {
			if (errno == EINTR) {
				pr_warn("%s: read returned -EINTR; retrying\n",
						__func__);
				continue;
			}
			res = -errno;
		}
		if ((res != size) && (errno == EINTR)) {
			pr_warn("%s: read interrupted: %ld < %ld; retrying\n",
					__func__, res, size);
			continue;
		}

		return res;
	}
	pr_warn("%s: failed; returning -EINTR\n", __func__);
	return -EINTR;
}

static int proxy_read_buf(const char *path, struct fuse_bufvec **bufp,
			size_t size, off_t offset, struct fuse_file_info *fi)
{
	struct fuse_bufvec *src;

	(void) path;

	src = malloc(sizeof(struct fuse_bufvec));
	if (src == NULL)
		return -ENOMEM;

	*src = FUSE_BUFVEC_INIT(size);

	src->buf[0].flags = FUSE_BUF_IS_FD | FUSE_BUF_FD_SEEK;
	src->buf[0].fd = fi->fh;
	src->buf[0].pos = offset;

	*bufp = src;

	return 0;
}

static int proxy_write(const char *path, const char *buf, size_t size,
		     off_t offset, struct fuse_file_info *fi)
{
	ssize_t res;
	int tries = 0;

	(void) path;

	while (tries++ < EINTR_TRIES) {
		errno = 0;
		res = pwrite(fi->fh, buf, size, offset);
		if (res == -1) {
			if (errno == EINTR) {
				pr_warn("%s: write returned -EINTR; retrying\n",
						__func__);
				continue;
			}
			res = -errno;
		}
		if ((res != size) && (errno == EINTR)) {
			pr_warn("%s: wrote interrupted: %ld < %ld; retrying\n",
					__func__, res, size);
			continue;
		}

		return res;
	}
	pr_warn("%s: failed; returning -EINTR\n", __func__);
	return -EINTR;
}

static int proxy_write_buf(const char *path, struct fuse_bufvec *buf,
		     off_t offset, struct fuse_file_info *fi)
{
	struct fuse_bufvec dst = FUSE_BUFVEC_INIT(fuse_buf_size(buf));

	(void) path;

	dst.buf[0].flags = FUSE_BUF_IS_FD | FUSE_BUF_FD_SEEK;
	dst.buf[0].fd = fi->fh;
	dst.buf[0].pos = offset;

	return fuse_buf_copy(&dst, buf, FUSE_BUF_SPLICE_NONBLOCK);
}

static int proxy_statfs(const char *path, struct statvfs *stbuf)
{
	int res;

	res = statvfs(path, stbuf);
	if (res == -1)
		return -errno;

	return 0;
}

static int proxy_flush(const char *path, struct fuse_file_info *fi)
{
	int res;

	(void) path;
	/* This is called from every close on an open file, so call the
	   close on the underlying filesystem.	But since flush may be
	   called multiple times for an open file, this must not really
	   close the file.  This is important if used on a network
	   filesystem like NFS which flush the ctx/metadata on close() */
	res = dup(fi->fh);
	if (res < 0)
		return -errno;
	res = close(res);
	if (res == -1)
		return -errno;

	return 0;
}

static int proxy_release(const char *path, struct fuse_file_info *fi)
{
	pr_debug("%s: closed %s as fd %d\n", __func__, path, fi->fh);

	(void) path;
	close(fi->fh);

	return 0;
}

static int proxy_fsync(const char *path, int isdatasync,
		     struct fuse_file_info *fi)
{
	int res;
	(void) path;

#ifndef HAVE_FDATASYNC
	(void) isdatasync;
#else
	if (isdatasync)
		res = fdatasync(fi->fh);
	else
#endif
		res = fsync(fi->fh);
	if (res == -1)
		return -errno;

	return 0;
}

#ifdef HAVE_POSIX_FALLOCATE
static int proxy_fallocate(const char *path, int mode,
			off_t offset, off_t length, struct fuse_file_info *fi)
{
	(void) path;

	if (mode)
		return -EOPNOTSUPP;

	return -posix_fallocate(fi->fh, offset, length);
}
#endif

#ifdef HAVE_SETXATTR
/* xattr operations are optional and can safely be left unimplemented */
static int proxy_setxattr(const char *path, const char *name, const char *value,
			size_t size, int flags)
{
	int res = lsetxattr(path, name, value, size, flags);
	if (res == -1)
		return -errno;
	return 0;
}

static int proxy_getxattr(const char *path, const char *name, char *value,
			size_t size)
{
	int res = lgetxattr(path, name, value, size);
	if (res == -1)
		return -errno;
	return res;
}

static int proxy_listxattr(const char *path, char *list, size_t size)
{
	int res = llistxattr(path, list, size);
	if (res == -1)
		return -errno;
	return res;
}

static int proxy_removexattr(const char *path, const char *name)
{
	int res = lremovexattr(path, name);
	if (res == -1)
		return -errno;
	return 0;
}
#endif /* HAVE_SETXATTR */

static int proxy_lock(const char *path, struct fuse_file_info *fi, int cmd,
		    struct flock *lock)
{
	(void) path;

	return ulockmgr_op(fi->fh, cmd, lock, &fi->lock_owner,
			   sizeof(fi->lock_owner));
}

static int proxy_flock(const char *path, struct fuse_file_info *fi, int op)
{
	int res;
	(void) path;

	res = flock(fi->fh, op);
	if (res == -1)
		return -errno;

	return 0;
}

struct fuse_operations proxy_operations = {
	.getattr	= proxy_getattr,
	.fgetattr	= proxy_fgetattr,
	.access		= proxy_access,
	.readlink	= proxy_readlink,
	.opendir	= proxy_opendir,
	.readdir	= proxy_readdir,
	.releasedir	= proxy_releasedir,
	.mknod		= proxy_mknod,
	.mkdir		= proxy_mkdir,
	.symlink	= proxy_symlink,
	.unlink		= proxy_unlink,
	.rmdir		= proxy_rmdir,
	.rename		= proxy_rename,
	.link		= proxy_link,
	.chmod		= proxy_chmod,
	.chown		= proxy_chown,
	.truncate	= proxy_truncate,
	.ftruncate	= proxy_ftruncate,
#ifdef HAVE_UTIMENSAT
	.utimens	= proxy_utimens,
#endif
	.create		= proxy_create,
	.open		= proxy_open,
	.read		= proxy_read,
	.read_buf	= proxy_read_buf,
	.write		= proxy_write,
	.write_buf	= proxy_write_buf,
	.statfs		= proxy_statfs,
	.flush		= proxy_flush,
	.release	= proxy_release,
	.fsync		= proxy_fsync,
#ifdef HAVE_POSIX_FALLOCATE
	.fallocate	= proxy_fallocate,
#endif
#ifdef HAVE_SETXATTR
	.setxattr	= proxy_setxattr,
	.getxattr	= proxy_getxattr,
	.listxattr	= proxy_listxattr,
	.removexattr	= proxy_removexattr,
#endif
	.lock		= proxy_lock,
	.flock		= proxy_flock,

	.flag_nullpath_ok = 1,
#if HAVE_UTIMENSAT
	.flag_utime_omit_ok = 1,
#endif
};
